#include "Tandberg.h"

int _showLog = 0;
HINSTANCE _hMyHandle;

#define SCT_LEN(sct) (3+((sct[1]&0x0f)<<8)+sct[2])

uint32_t b2i(int32_t n, const uint8_t *b)
{
    switch(n)
    {
    case 2:
        return  (b[0] <<  8) |  b[1];
    case 3:
        return  (b[0] << 16) | (b[1] <<  8) |  b[2];
    case 4:
        return ((b[0] << 24) | (b[1] << 16) | (b[2] <<8 ) | b[3]) & 0xffffffff;
    default:
        return 0;
    }
    return 0;
}

char * Version()
{
    return
        "V 1.3";
}

char * Author()
{
    return
        "SPC";
}

bool CheckCurrentCAID(DWORD CAID)
{
    return (CAID == 0x1010);
}

bool ECMFunction(byte * data, byte * dw)
{
    uint8_t g_ecmkey[8] = { 0 };

    memset(dw, 0, sizeof(dw));

    uint16_t ecmLen = SCT_LEN(data);
    if (_showLog) SendMSG(0, "EcmLen: %d", ecmLen);
    if (_showLog) SendHexDump("Received ECM: ", 0, data, ecmLen + 3);

    byte mode = data[3];
    if (_showLog) SendMSG(0, "Mode: 0x%X", mode);


    if(ecmLen < 5)
    {
        SendMSG(0, "bad ecmLen");
        return false;
    }

    uint32_t pos = 3;
    do
    {
        uint8_t nano_type = data[pos];
        uint8_t nano_length = data[pos+1];

        if(pos + 2 + nano_length > ecmLen)
        {
            SendMSG(0, "bad ecmLen: %d > %d", pos + 2 + nano_length, ecmLen);
            break;
        }

        uint8_t* nano_data = data + pos + 2;

        switch(nano_type)
        {

        case 0xED: // ECM_TAG_CW_DESCRIPTOR
        {
            if(nano_length != 0x26)
            {
                SendMSG(0, "warning: ECM_TAG_CW_DESCRIPTOR length (%d) != %d", nano_length, 0x26);
                break;
            }

            uint32_t entitlement_id = b2i(4, nano_data);
            RequestKey("T", entitlement_id, "01", g_ecmkey);
            if (IsKeyNull(g_ecmkey, 8))
            {
                SendMSG(1,"T %X Key%s is missing...", entitlement_id, "01");
                return false;
            }

            byte encryptedData[32] = { 0 };
            memcpy(encryptedData, nano_data + 4, 32);
            if (_showLog) SendHexDump("Encrypted Data: ", 0, encryptedData, 32);

            byte iv[8] = { 0 };
            des_cbc_decrypt(encryptedData, iv, g_ecmkey, 32);
            if (_showLog) SendHexDump("Decrypted Data: ", 0, encryptedData, 32);

            int offset = 8;
            dw[0 + offset] = encryptedData[0x16];
            dw[1 + offset] = encryptedData[0x03];
            dw[2 + offset] = encryptedData[0x11];
            dw[3 + offset] = (dw[0 + offset] + dw[1 + offset] + dw[2 + offset]) & 0xff;
            dw[4 + offset] = encryptedData[0x18];
            dw[5 + offset] = encryptedData[0x10];
            dw[6 + offset] = encryptedData[0x0E];
            dw[7 + offset] = (dw[4 + offset] + dw[5 + offset] + dw[6 + offset]) & 0xff;

            offset = 0;
            dw[0 + offset] = encryptedData[0x05];
            dw[1 + offset] = encryptedData[0x19];
            dw[2 + offset] = encryptedData[0x1D];
            dw[3 + offset] = (dw[0 + offset] + dw[1 + offset] + dw[2 + offset]) & 0xff;
            dw[4 + offset] = encryptedData[0x0B];
            dw[5 + offset] = encryptedData[0x12];
            dw[6 + offset] = encryptedData[0x1A];
            dw[7 + offset] = (dw[4 + offset] + dw[5 + offset] + dw[6 + offset]) & 0xff;

            return true;
        }

        case 0xEE: // ECM_TAG_CW_DESCRIPTOR
        {
            if(nano_length != 0x16)
            {
                SendMSG(0, "warning: ECM_TAG_CW_DESCRIPTOR length (%d) != %d", nano_length, 0x16);
                break;
            }

            uint32_t entitlement_id = b2i(4, nano_data);
            RequestKey("T", entitlement_id, "01", g_ecmkey);
            if (IsKeyNull(g_ecmkey, 8))
            {
                SendMSG(1,"T %X Key%s is missing...", entitlement_id, "01");
                return false;
            }

            uint32_t ks[32] = { 0 };
            des_set_key(g_ecmkey, ks);

            memcpy(dw, nano_data + 4 + 8, 8); // even
            memcpy(dw + 8, nano_data + 4, 8); // odd

            des(dw, ks, 0);
            des(dw + 8, ks, 0);

            return true;
        }

        case 0xEC:
        {
            uint8_t mode = nano_data[5];
            uint8_t dataCrypted[0x20] = { 0 };
            memcpy(dataCrypted, nano_data + 6, 0x20);
            if (_showLog) SendHexDump("dataCrypted: ", 0, dataCrypted, 32);

            //Step 1 - Decrypt DES CBC with normal Key and IV 0000000000000000 (equal to nanoED)
            uint32_t entitlement_id = b2i(4, nano_data);
            RequestKey("T", entitlement_id, "01", g_ecmkey);
            if (IsKeyNull(g_ecmkey, 8))
            {
                SendMSG(1,"T %X Key%s is missing...", entitlement_id, "01");
                return false;
            }

            byte iv[8] = { 0 };
            des_cbc_decrypt(dataCrypted, iv, g_ecmkey, 32);
            if (_showLog) SendHexDump("Decrypted Data: ", 0, dataCrypted, 32);

            if(_showLog) SendMSG(0, "nanoEC, mode: %X, IsLikeNanoED: %d", mode, (mode & 0x20) == 0);

            if ((mode & 0x20) == 0)
            {
                //Step 2 - Create CW (equal to nano ED)
                int offset = 8;
                dw[0 + offset] = dataCrypted[0x16];
                dw[1 + offset] = dataCrypted[0x03];
                dw[2 + offset] = dataCrypted[0x11];
                dw[3 + offset] = (dw[0 + offset] + dw[1 + offset] + dw[2 + offset]) & 0xff;
                dw[4 + offset] = dataCrypted[0x18];
                dw[5 + offset] = dataCrypted[0x10];
                dw[6 + offset] = dataCrypted[0x0E];
                dw[7 + offset] = (dw[4 + offset] + dw[5 + offset] + dw[6 + offset]) & 0xff;

                offset = 0;
                dw[0 + offset] = dataCrypted[0x05];
                dw[1 + offset] = dataCrypted[0x19];
                dw[2 + offset] = dataCrypted[0x1D];
                dw[3 + offset] = (dw[0 + offset] + dw[1 + offset] + dw[2 + offset]) & 0xff;
                dw[4 + offset] = dataCrypted[0x0B];
                dw[5 + offset] = dataCrypted[0x12];
                dw[6 + offset] = dataCrypted[0x1A];
                dw[7 + offset] = (dw[4 + offset] + dw[5 + offset] + dw[6 + offset]) & 0xff;

            }
            else //new algo (overencryption with AES)
            {
                //Step 2 - Prepare data for AES (it is like the creation of CW in nanoED but swapped each 8 bytes)
                uint8_t dataEC[16] = { 0 };
                dataEC[7] = dataCrypted[0x16];
                dataEC[6] = dataCrypted[0x03];
                dataEC[5] = dataCrypted[0x11];
                dataEC[4] = dataCrypted[0x09];
                dataEC[3] = dataCrypted[0x18];
                dataEC[2] = dataCrypted[0x10];
                dataEC[1] = dataCrypted[0x0E];
                dataEC[0] = dataCrypted[0x02];

                dataEC[15] = dataCrypted[0x05];
                dataEC[14] = dataCrypted[0x19];
                dataEC[13] = dataCrypted[0x1D];
                dataEC[12] = dataCrypted[0x04];
                dataEC[11] = dataCrypted[0x0B];
                dataEC[10] = dataCrypted[0x12];
                dataEC[9] = dataCrypted[0x1A];
                dataEC[8] = dataCrypted[0x13];

                //Step 3 - Decrypt AES CBC with new Aes Key and IV 2EBD816A5E749A708AE45ADDD84333DE
                byte keySelectAesKey = mode & 0x1F; //32 possible AES keys

                char keynr_str[4];
                IntToHex(keynr_str, keySelectAesKey);

                uint8_t keyAes[16] = { 0 };
                RequestKey("T", 1, keynr_str, keyAes);
                if (IsKeyNull(keyAes, 16))
                {
                    SendMSG(1,"T 0000 Key%s is missing... (AES Key)", keynr_str);
                    return false;
                }

                if (_showLog) SendHexDump("Using AesKey: ", 0, keyAes, 16);

                byte ivEC[16] = { 0x2E, 0xBD, 0x81, 0x6A, 0x5E, 0x74, 0x9A, 0x70, 0x8A, 0xE4, 0x5A, 0xDD, 0xD8, 0x43, 0x33, 0xDE };
                AES_KEY dkey;
                AES_set_decrypt_key(keyAes, 128, &dkey);
                AES_cbc_encrypt(dataEC, dataEC, 16, (const AES_KEY *)&dkey, ivEC, AES_DECRYPT);
                if (_showLog) SendHexDump("Final dataEC: ", 0, dataEC, 16);

                //Step 4 - Create CW (a simple swap)
                int offset = 8;
                dw[7 + offset] = dataEC[0];
                dw[6 + offset] = dataEC[1];
                dw[5 + offset] = dataEC[2];
                dw[4 + offset] = dataEC[3];
                dw[3 + offset] = dataEC[4];
                dw[2 + offset] = dataEC[5];
                dw[1 + offset] = dataEC[6];
                dw[0 + offset] = dataEC[7];
                dw[3 + offset] = (dw[0 + offset] + dw[1 + offset] + dw[2 + offset]) & 0xff;
                dw[7 + offset] = (dw[4 + offset] + dw[5 + offset] + dw[6 + offset]) & 0xff;

                offset = 0;
                dw[7 + offset] = dataEC[8];
                dw[6 + offset] = dataEC[9];
                dw[5 + offset] = dataEC[10];
                dw[4 + offset] = dataEC[11];
                dw[3 + offset] = dataEC[12];
                dw[2 + offset] = dataEC[13];
                dw[1 + offset] = dataEC[14];
                dw[0 + offset] = dataEC[15];
                dw[3 + offset] = (dw[0 + offset] + dw[1 + offset] + dw[2 + offset]) & 0xff;
                dw[7 + offset] = (dw[4 + offset] + dw[5 + offset] + dw[6 + offset]) & 0xff;
            }

            return true;
        }

        default:
            break;
        }

        pos += 2 + nano_length;

    }
    while (pos < ecmLen);

    return false;
}

void SetShowDebug(int showLog)
{
    _showLog = showLog;
}

bool EMMFunction(byte * data)
{
    return false;
}

//it's a fix for mingw!
//http://forums.codeblocks.org/index.php/topic,946.0.html
#ifdef __cplusplus
extern "C"
#endif

BOOL WINAPI DllMain( HINSTANCE hModule,
                     DWORD  ul_reason_for_call,
                     LPVOID lpReserved)
{
    switch (ul_reason_for_call)
    {
    case DLL_PROCESS_ATTACH:
        // attach to process
        // return FALSE to fail DLL load
        _hMyHandle = hModule;
        break;

    case DLL_PROCESS_DETACH:
        // detach from process
        break;

    case DLL_THREAD_ATTACH:
        // attach to thread
        break;

    case DLL_THREAD_DETACH:
        // detach from thread
        break;
    }
    return TRUE; // succesful

}
